/*
 * robot_impl.c
 *
 *  Created on: 20.07.2016
 *      Author: AndreR
 */

#include "main/robot.h"
#include "main/ctrl.h"
#include "util/log_file.h"
#include "hal/motors.h"
#include "hal/kicker.h"
#include "hal/led.h"
#include "spi4.h"
#include "spi6.h"
#include "power.h"
#include "util/sys_time.h"
#include "main/network.h"
#include "hal/buzzer.h"
#include "ext.h"

BotParameters botParams = {
	.driveTrain = {
		.motor2WheelRatio = (18.0f/60.0f),
		.wheel2MotorRatio = (60.0f/18.0f),
		.motor = {
			.Km = 0.0251f,
			.Kn = (380.0f*2.0f*PI/60.0f),
			.Ke = (60.0f/(380.0f*2.0f*PI)),
			.Kf = 0.94e-4f,
			.R = 0.464f,
			.L = 322e-6f,
		}
	},
	.physical = { 0.0285f, 30.0f, 45.0f, 0.083f, 2.3f },
	.v2016 = 1,
};

void RobotImplUpdateSensors(RobotSensors* pSensors)
{
	// motors
	pSensors->enc.time = motors.drive[0].velocityTimestamp;
	pSensors->motorVol.time = motors.drive[0].outVoltageTimestamp;

	for(uint8_t m = 0; m < 4; m++)
	{
		pSensors->enc.vel[m] = motors.drive[m].velocity;
		pSensors->motorVol.vol[m] = motors.drive[m].outVoltage;
	}

	// motor current
	pSensors->motorCur.time = spi4.motors.measTimestamp;
	pSensors->motorCur.cur[0] = -spi4.motors.current[1];
	pSensors->motorCur.cur[1] = -spi4.motors.current[0];
	pSensors->motorCur.cur[2] = -spi4.motors.current[3];
	pSensors->motorCur.cur[3] =  spi4.motors.current[2];

	// power
	pSensors->power.voltage = power.vBat;
	pSensors->power.current = power.iCur;
	pSensors->power.batEmpty = power.batEmpty;
	pSensors->power.exhausted = power.exhausted;

	// IMU
	pSensors->acc.linAcc[0] = spi6.imu.acc[1];
	pSensors->acc.linAcc[1] = -spi6.imu.acc[0];
	pSensors->acc.linAcc[2] = spi6.imu.acc[2];
	pSensors->acc.time = SysTimeUSec();
	pSensors->acc.updated = 1;

	pSensors->gyr.rotVel[0] = spi6.imu.gyr[1];
	pSensors->gyr.rotVel[1] = -spi6.imu.gyr[0];
	pSensors->gyr.rotVel[2] = spi6.imu.gyr[2];
	pSensors->gyr.time = SysTimeUSec();
	pSensors->gyr.updated = 1;

	// Vision - SKIPPED (filled by robot base implementation)

	// Kicker
	pSensors->kicker.level = kicker.vCap;
	pSensors->kicker.chgCurrent = kicker.iChg;
	pSensors->kicker.kickCounter = kicker.kickCounter;
	pSensors->kicker.flags = 0;
	if(KickerIsReadyToShoot())
		pSensors->kicker.flags |= KICKER_FLAGS_READY;

	if(kicker.barrierDamaged)
		pSensors->kicker.flags |= KICKER_FLAGS_DMG_BARRIER;

	if(kicker.straightDamaged)
		pSensors->kicker.flags |= KICKER_FLAGS_DMG_STRAIGHT;

	if(kicker.chipDamaged)
		pSensors->kicker.flags |= KICKER_FLAGS_DMG_CHIP;

	if(kicker.chargeOverheated)
		pSensors->kicker.flags |= KICKER_FLAGS_DMG_CHARGE;

	if(kicker.v2017)
		pSensors->kicker.flags |= KICKER_FLAGS_V2017;

	// Dribbler
	pSensors->dribbler.speed = motors.dribbler.velocity*MOTORS_DRIBBLER_MOTOR_RADS_TO_BAR_RPM;
	pSensors->dribbler.voltage = motors.dribbler.outVoltage;
//	pSensors->dribbler.temp = 0;	// TODO: implement for new motor
	pSensors->dribbler.temp = kicker.tEnv;
	pSensors->dribbler.overheated = 0;

	// Barrier
	pSensors->ir.level[0] = kicker.vIrOff;
	pSensors->ir.level[1] = kicker.vIrOn;
	pSensors->ir.interrupted = kicker.interrupted;
}

void RobotImplWriteLogMsg(void* pMsg)
{
	LogFileWrite(pMsg);
}

void RobotImplMotorOutput(const RobotCtrlOutput* pMotor)
{
	motors.driveMode = pMotor->ctrlMode;

	for(uint8_t i = 0; i < 4; i++)
	{
		motors.drive[i].setVelocity = pMotor->motorVel[i];
		motors.drive[i].setVoltage = pMotor->motorVoltage[i];
		motors.drive[i].setAcceleration = pMotor->motorAcc[i];
		motors.drive[i].setTorque = pMotor->motorTorque[i];
	}
}

void RobotImplKickerOutput(const KickerInput* pKicker)
{
	switch(pKicker->mode)
	{
	case KICKER_MODE_FORCE:
		KickerFireSpeed(pKicker->speed, pKicker->device);
		break;
	case KICKER_MODE_ARM:
		KickerArmSpeed(pKicker->speed, pKicker->device);
		break;
	case KICKER_MODE_DISARM:
		KickerDisarm();
		break;
	case KICKER_MODE_ARM_TIME:
		KickerArm(pKicker->speed, pKicker->device);
		break;
	}
}

void RobotImplDribblerOutput(const DribblerInput* pDribbler)
{
	switch(pDribbler->mode)
	{
	case DRIBBLER_MODE_VOLTAGE:
		MotorsDribblerSetVoltage(pDribbler->voltage);
		break;
	case DRIBBLER_MODE_SPEED:
		MotorsDribblerSetVelocity(pDribbler->speed);
		break;
	default:
		MotorsDribblerOff();
		break;
	}
}

void RobotImplBuzzerPlay(uint16_t seqId)
{
	BuzzerPlayId(seqId);
}

void RobotImplKickerAutoCharge(uint8_t enable)
{
	if(enable)
	{
		KickerEnableAutoRecharge(1);
		KickerEnableCharge(1);
	}
	else
	{
		KickerAutoDischarge();
	}
}

void RobotImplSetLEDs(uint8_t state)
{
	LEDSet(state);
}

void RobotImplExtSendMatchFeedback(const SystemMatchFeedback* pFeedback)
{
	PacketHeader header;
	header.section = SECTION_SYSTEM;
	header.cmd = CMD_SYSTEM_MATCH_FEEDBACK;
	ExtSendPacket(&header, pFeedback, sizeof(SystemMatchFeedback));
}
