/*
 * prog_store.c
 *
 *  Created on: 01.04.2015
 *      Author: AndreR
 */

#include "prog_store.h"
#include "util/intercom_reliable.h"
#include "util/console.h"
#include "util/crc.h"
#include "util/boot.h"
#include "util/log.h"
#include "ch.h"

#include <string.h>

#define PROG_STORE_EVENT_TYPE_ERROR	0
#define PROG_STORE_EVENT_TYPE_DATA	1
#define PROG_STORE_EVENT_TYPE_SIZE	2
#define PROG_STORE_EVENT_TYPE_CRC32	3

static THD_WORKING_AREA(waProgStore, 8192);

ProgStore progStore;

static void bootloaderCb(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)dataLength;
	(void)pUserData;

	switch(pHeader->command)
	{
		case BOOTLOADER_NACK:
		{
			BootloaderSeq* pSeq = (BootloaderSeq*)pData;

			if(pSeq->seq == progStore.expectedSeq)
			{
				progStore.progAckd = 0;
				chBSemSignal(&progStore.progSem);
			}
		}
		break;
		case BOOTLOADER_ACK:
		{
			BootloaderSeq* pSeq = (BootloaderSeq*)pData;

			if(pSeq->seq == progStore.expectedSeq)
			{
				progStore.progAckd = 1;
				chBSemSignal(&progStore.progSem);
			}
		}
		break;
	}
}

static uint8_t progClientErase(uint8_t index)
{
	progStore.expectedSeq = progStore.progSeq;
	BootloaderSeq seq = {progStore.progSeq++};

	IntercomHeader header;
	header.address = progStore.targets[index].intercomAddress;
	header.command = BOOTLOADER_ERASE;
	header.section = INTERCOM_SECTION_BOOTLOADER;

	progStore.progAckd = 0;
	chBSemReset(&progStore.progSem, 1);

	do
	{
		IntercomSend(&header, sizeof(BootloaderSeq), &seq);
	}
	while(chBSemWaitTimeout(&progStore.progSem, MS2ST(1000)) != MSG_OK);

	return progStore.progAckd;
}

static uint8_t progClientCrc(uint8_t index, uint32_t start, uint32_t end, uint32_t expectedCRC)
{
	progStore.expectedSeq = progStore.progSeq;
	BootloaderChecksum crc;
	crc.seq = progStore.progSeq++;
	crc.start = start;
	crc.end = end;
	crc.crc = expectedCRC;

	IntercomHeader header;
	header.address = progStore.targets[index].intercomAddress;
	header.command = BOOTLOADER_CHECKSUM;
	header.section = INTERCOM_SECTION_BOOTLOADER;

	progStore.progAckd = 0;
	chBSemReset(&progStore.progSem, 1);

	do
	{
		IntercomSend(&header, sizeof(BootloaderChecksum), &crc);
	}
	while(chBSemWaitTimeout(&progStore.progSem, MS2ST(100)) != MSG_OK);

	return progStore.progAckd;
}

static uint8_t progClientData(uint8_t index, uint32_t offset, uint32_t dataLength)
{
	ProgStoreImplReadData(index, offset, progStore.progDataBuf+sizeof(BootloaderProgramData), dataLength);

	progStore.expectedSeq = progStore.progSeq;

	BootloaderProgramData* pProgData = (BootloaderProgramData*)progStore.progDataBuf;
	pProgData->seq = progStore.progSeq++;
	pProgData->offset = offset;

	IntercomHeader header;
	header.address = progStore.targets[index].intercomAddress;
	header.command = BOOTLOADER_DATA;
	header.section = INTERCOM_SECTION_BOOTLOADER;

	progStore.progAckd = 0;
	chBSemReset(&progStore.progSem, 1);

	do
	{
		IntercomSend(&header, dataLength+sizeof(BootloaderProgramData), progStore.progDataBuf);
	}
	while(chBSemWaitTimeout(&progStore.progSem, MS2ST(200)) != MSG_OK);

	return progStore.progAckd;
}

static void progClientSelectBootloader()
{
	IntercomHeader iHeader;
	iHeader.section = INTERCOM_SECTION_BOOTLOADER;
	iHeader.address = 0;
	iHeader.command = BOOTLOADER_SELECT;

	for(uint8_t i = 0; i < progStore.usedTargets; i++)
	{
		iHeader.address |= progStore.targets[i].intercomAddress;
	}

	IntercomReliableSend(&iHeader, 0, 0);
}

static uint8_t programId2Index(uint8_t programId)
{
	for(uint8_t i = 0; i < progStore.usedTargets; i++)
	{
		if(progStore.targets[i].programId == programId)
			return i;
	}

	return 0xFF;
}

void ProgStoreInit()
{
	chBSemObjectInit(&progStore.progSem, 0);
	chBSemObjectInit(&progStore.storeSem, 0);

	progStore.expectedSeq = 1;

	IntercomClient iClient;
	iClient.pUserData = 0;
	iClient.section = INTERCOM_SECTION_BOOTLOADER;
	iClient.rxCallback = &bootloaderCb;

	IntercomAddClient(iClient);

	ProgStoreImplInit();
}

void ProgStoreProvideError(uint8_t procId, int16_t error)
{
	ConsolePrint("Error: %hd\r\n", error);

	uint8_t index = programId2Index(procId);

	// e.g. media detached, link lost
	if(index >= progStore.usedTargets)
		return;

	progStore.lastEvent.type = PROG_STORE_EVENT_TYPE_ERROR;
	progStore.lastEvent.procId = index;
	progStore.lastEvent.errorId = error;

	chBSemSignal(&progStore.storeSem);
}

void ProgStoreProvideData(uint8_t procId, uint32_t offset, uint8_t* pData, uint32_t size)
{
	uint8_t index = programId2Index(procId);

	if(index >= progStore.usedTargets)
		return;

	if(offset+size > progStore.targets[index].flashSize)
		return;

	ProgStoreImplWriteData(index, offset, pData, size);

	progStore.lastEvent.type = PROG_STORE_EVENT_TYPE_DATA;
	progStore.lastEvent.procId = index;
	progStore.lastEvent.data.offset = offset;
	progStore.lastEvent.data.size = size;

	chBSemSignal(&progStore.storeSem);
}

void ProgStoreProvideSize(uint8_t procId, uint32_t size)
{
	uint8_t index = programId2Index(procId);

	if(index >= progStore.usedTargets)
		return;

	progStore.lastEvent.type = PROG_STORE_EVENT_TYPE_SIZE;
	progStore.lastEvent.procId = index;
	progStore.lastEvent.size = size;

	chBSemSignal(&progStore.storeSem);
}

void ProgStoreProvideCRC32(uint8_t procId, uint32_t startAddr, uint32_t endAddr, uint32_t crc32)
{
	uint8_t index = programId2Index(procId);

	if(index >= progStore.usedTargets)
		return;

	progStore.lastEvent.type = PROG_STORE_EVENT_TYPE_CRC32;
	progStore.lastEvent.procId = index;
	progStore.lastEvent.crc32.startAddr = startAddr;
	progStore.lastEvent.crc32.endAddr = endAddr;
	progStore.lastEvent.crc32.crc = crc32;

	chBSemSignal(&progStore.storeSem);
}

static int16_t waitForEvent(ProgStoreEvent* pEventIn, uint32_t timeoutMs)
{
	int16_t result = 0;
	ProgStoreEvent* pCurEvent = &progStore.lastEvent;

	while(1)
	{
		if(chBSemWaitTimeout(&progStore.storeSem, MS2ST(timeoutMs)) != MSG_OK)
			return PROG_STORE_ERROR_TIMEOUT;

		if(pCurEvent->type == PROG_STORE_EVENT_TYPE_ERROR)
		{
			result = pCurEvent->errorId;
			break;
		}

		if(pCurEvent->type == pEventIn->type && pCurEvent->procId == pEventIn->procId)
		{
			if(pCurEvent->type == PROG_STORE_EVENT_TYPE_SIZE)
			{
				pEventIn->size = pCurEvent->size;
				break;
			}

			if(	pCurEvent->type == PROG_STORE_EVENT_TYPE_DATA &&
				pEventIn->data.offset == pCurEvent->data.offset)
			{
				pEventIn->data.size = pCurEvent->data.size;
				break;
			}

			if(	pCurEvent->type == PROG_STORE_EVENT_TYPE_CRC32 &&
				pEventIn->crc32.startAddr == pCurEvent->crc32.startAddr &&
				pEventIn->crc32.endAddr == pCurEvent->crc32.endAddr)
			{
				pEventIn->crc32.crc = pCurEvent->crc32.crc;
				break;
			}
		}
	}

	return result;
}

static int16_t waitForData(ProgSource* pSource, uint8_t index, uint32_t offset, uint32_t size)
{
	int16_t result;
	ProgStoreEvent ev;

	ev.procId = index;
	ev.type = PROG_STORE_EVENT_TYPE_DATA;
	ev.data.offset = offset;

	while(1)
	{
		chBSemReset(&progStore.storeSem, 1);

		(*pSource->pDataFunc)(progStore.targets[index].programId, offset, size);

		result = waitForEvent(&ev, pSource->timeout);
		if(result == PROG_STORE_ERROR_TIMEOUT)
			continue;

		if(ev.data.size != size)
			continue;

		if(result)
			return result;

		break;
	}

	return 0;
}

static int16_t waitForSize(ProgSource* pSource, int8_t index, uint32_t* pSize)
{
	int16_t result;
	ProgStoreEvent ev;

	ev.procId = index;
	ev.type = PROG_STORE_EVENT_TYPE_SIZE;

	while(1)
	{
		chBSemReset(&progStore.storeSem, 1);

		(*pSource->pSizeFunc)(progStore.targets[index].programId);

		result = waitForEvent(&ev, pSource->timeout);
		if(result == PROG_STORE_ERROR_TIMEOUT)
			continue;

		if(result)
			return result;

		break;
	}

	if(pSize)
		*pSize = ev.size;

	return 0;
}

static int16_t waitForCrc(ProgSource* pSource, int8_t index, uint32_t startAddr, uint32_t endAddr, uint32_t* pCRC)
{
	int16_t result;
	ProgStoreEvent ev;

	ev.procId = index;
	ev.type = PROG_STORE_EVENT_TYPE_CRC32;
	ev.crc32.startAddr = startAddr;
	ev.crc32.endAddr = endAddr;

	while(1)
	{
		chBSemReset(&progStore.storeSem, 1);

		(*pSource->pCRC32Func)(progStore.targets[index].programId, startAddr, endAddr);

		result = waitForEvent(&ev, pSource->timeout);
		if(result == PROG_STORE_ERROR_TIMEOUT)
			continue;

		if(result)
			return result;

		break;
	}

	if(pCRC)
		*pCRC = ev.crc32.crc;

	return 0;
}

void ProgFlashTask(void* params)
{
	(void)params;

	for(uint16_t index = 0; index < progStore.usedTargets; index++)
	{
		ProgUpgradeInfo upgradeInfo;
		ProgStoreImplGetUpgradeInfo(index, &upgradeInfo);

		if(upgradeInfo.update == 0)
			continue;

		progClientErase(index);

		for(uint32_t offset = 0; offset < upgradeInfo.size; offset += 128)
		{
			if(progClientData(index, offset, 128) == 0)
			{
				offset -= 128;
			}

			uint32_t crc = ProgStoreImplGetChecksum(index, offset, 128);

			if(progClientCrc(index, offset, offset+128, crc) == 0)
			{
				offset -= 128;
			}
		}

		uint32_t crc = ProgStoreImplGetChecksum(index, 0, progStore.targets[index].flashSize);

		if(progClientCrc(index, 0, progStore.targets[index].flashSize, crc) == 0)
		{
			index--;
		}
	}

	BootReset();
}

void ProgStoreTask(void* params)
{
	(void)params;
	int16_t result;
	uint8_t sourceAvail;

	chRegSetThreadName("ProgStore");

	// 1. get program size for main to check source availability
	(*progStore.sources[progStore.useSource].pOpenFunc)();

	result = waitForSize(&progStore.sources[progStore.useSource], 0, 0);
	if(result)
		sourceAvail = 0;
	else
		sourceAvail = 1;

	(*progStore.sources[progStore.useSource].pCloseFunc)();

	// 2. select prog source
	if(!sourceAvail)
	{
		ConsolePrint("No program source available\r\n");
		chThdExit(0);
	}

	// erase storage
	ProgStoreImplEraseStorage();

	// 3. get CRCs and sizes from selected prog source
	ProgSource* pSource = &progStore.sources[progStore.useSource];

	(*pSource->pOpenFunc)();

	for(uint16_t proc = 0; proc < progStore.usedTargets; proc++)
	{
		ConsolePrint("Calculating proc %hu CRC\r\n", proc);

		ProgUpgradeInfo upgradeInfo;
		upgradeInfo.crc = 0;
		upgradeInfo.size = 0;
		upgradeInfo.update = 0;

		result = waitForCrc(pSource, proc, 0, progStore.targets[proc].flashSize, &upgradeInfo.crc);
		if(result)
		{
			ConsolePrint("waitForCrc: %hd\r\n", result);
			chThdExit(0);
		}

		ConsolePrint("Proc %hu CRC: %08X\r\n", proc, upgradeInfo.crc);

		// skip processors with that flash content
		if(progClientCrc(proc, 0, progStore.targets[proc].flashSize, upgradeInfo.crc))
		{
			ProgStoreImplSetUpgradeInfo(proc, &upgradeInfo);

			ConsolePrint("CRC matched\r\n");
			continue;
		}

		ConsolePrint("No CRC match\r\n");

		upgradeInfo.update = 1;

		result = waitForSize(pSource, proc, &upgradeInfo.size);
		if(result)
			chThdExit(0);

		ConsolePrint("Proc %hu size: %u\r\n", proc, upgradeInfo.size);

		ProgStoreImplSetUpgradeInfo(proc, &upgradeInfo);
	}

	// 4. load prog source binaries to ext. RAM
	for(uint16_t proc = 0; proc < progStore.usedTargets; proc++)
	{
		ProgUpgradeInfo upgradeInfo;
		ProgStoreImplGetUpgradeInfo(proc, &upgradeInfo);

		if(upgradeInfo.update == 0)
			continue;

		ConsolePrint("Load %hu\r\n", proc);

		for(uint32_t offset = 0; offset < upgradeInfo.size; offset += pSource->maxTransferSize)
		{
			if(offset > 0 && offset%(10*pSource->maxTransferSize) == 0)
			{
				uint32_t tmpCrc;
				uint32_t startAddr = offset-10*pSource->maxTransferSize;
				result = waitForCrc(pSource, proc, startAddr, offset, &tmpCrc);

				if(ProgStoreImplGetChecksum(proc, startAddr, offset-startAddr) != tmpCrc)
				{
					ConsolePrint("Intermediate CRC check failure at %u of %hu\r\n", offset, proc);
					offset -= 11*pSource->maxTransferSize;
					continue;
				}
			}

			result = waitForData(pSource, proc, offset, pSource->maxTransferSize);
			if(result)
				chThdExit(0);
		}

		// verify checksum
		if(ProgStoreImplGetChecksum(proc, 0, progStore.targets[proc].flashSize) != upgradeInfo.crc)
		{
			ConsolePrint("Final CRC check failure for proc %hu\r\n", proc);
			--proc;
		}

		ConsolePrint("Proc %hu loaded\r\n", proc);
	}

	(*pSource->pCloseFunc)();

	uint16_t proc;
	for(proc = 0; proc < progStore.usedTargets; proc++)
	{
		ProgUpgradeInfo upgradeInfo;
		ProgStoreImplGetUpgradeInfo(proc, &upgradeInfo);

		if(upgradeInfo.update)
			break;
	}

	if(proc == progStore.usedTargets)
	{
		ConsolePrint("up to date\r\n");
		chThdExit(0);
	}

	for(uint16_t i = 0; i < progStore.usedTargets; i++)
	{
		ProgUpgradeInfo upgradeInfo;
		ProgStoreImplGetUpgradeInfo(i, &upgradeInfo);

		ConsolePrint("ID: %hu\r\nUpdate: %hu\r\n", i, (uint16_t)upgradeInfo.update);
		ConsolePrint("Size: %u\r\nCRC: 0x%08X\r\n", upgradeInfo.size, upgradeInfo.crc);
	}

	// 6. reboot & flash
	progClientSelectBootloader();
	BootReset();
}

void ProgStoreFlash()
{
	chThdCreateStatic(waProgStore, sizeof(waProgStore), NORMALPRIO-20, &ProgFlashTask, 0);
}

void ProgStoreStart(uint8_t source)
{
	progStore.useSource = source;

	if(progStore.pProgThread != 0)
	{
		if(chThdTerminatedX(progStore.pProgThread) == 0)
		{
			ConsolePrint("Program update already in progress\r\n");
			return;
		}
	}

	progStore.pProgThread = chThdCreateStatic(waProgStore, sizeof(waProgStore), NORMALPRIO-20, &ProgStoreTask, 0);
}

int16_t ProgStoreAddSource(const ProgSource* pSource)
{
	if(progStore.usedSources >= PROG_STORE_MAX_SOURCES)
		return -1;

	int16_t id = progStore.usedSources;

	memcpy(&progStore.sources[progStore.usedSources++], pSource, sizeof(ProgSource));

	return id;
}

int16_t ProgStoreAddTarget(uint8_t address, uint32_t flashSize, uint8_t programId)
{
	if(progStore.usedTargets >= PROG_STORE_MAX_TARGETS)
	{
		LogError("Max prog targets reached");
		return -1;
	}

	ProgTarget* pTarget = &progStore.targets[progStore.usedTargets];
	pTarget->intercomAddress = address;
	pTarget->flashSize = flashSize;
	pTarget->programId = programId;

	int16_t id = progStore.usedTargets;

	++progStore.usedTargets;

	return id;
}
